#SETUP TOEPLITZ STRUCTURE, FACTOR MODEL ESTIMATION BY POET-markowitz
#Tables 2a-b, POET
library(POET)
library(MASS)
library(mvtnorm)
set.seed(10)

rho1=0.000376#daily return that corresponds to yearly return

rhoz=0.15#correlation
n=100#sample size
p=0.5*n# of assets in portfolio
mu=rnorm(p)/100
omz=diag(p)#WILL FORM SIGMA

for(j in (1:p)){
  
  for (i in (1:p)){
    omz[i,j]=(rhoz)^{abs(i-j)}}
}
iomz=solve(omz)#very sparse, precision matrix, SIGMA^{-1} OR THETA
ones=rep(1,p)
ones=matrix(ones,p,1)

A=t(ones)%*%iomz%*%ones/p
B=t(ones)%*%iomz%*%mu/p
D=t(mu)%*%iomz%*%mu/p

w1=(D- rho1*B)/(A*D- B^2)
w1=matrix(w1,1,1)

w11=(iomz%*%ones/p)%*%w1
w2=(rho1*A - B)/(A*D - B^2)
w2=matrix(w2,1,1)
w21=(iomz%*%mu/p)%*%w2
ws=w11+w21#weight in markowitz
opv=(A*rho1^2-2*B*rho1+D)/(A*D-B^2)


#####
#GENERATE EXCESS ASSET RETURNS

ite=100
l1e=matrix(0,ite,1)
err1=matrix(0,ite,1)
thm3er=matrix(0,ite,1)
for (jj in 1:ite){
  #rr11=rnorm(n*p)
  #rr11=matrix(rr11,n,p)
  #sqomz=chol(omz)
  #r=rr11%*%sqomz
  r=rmvnorm(n*p,mu,omz)
  r=matrix(r,n,p)
  
  #generated
  ndmr=matrix(0,n,p)
  #form sigmahat
  rbar=colMeans(r)#should be p by 1
  rbar=matrix(rbar,p,1)
  for (t in 1:n)
  {ndmr[t,]=r[t,]-t(rbar)
  
  }
  ####
  
  covY=POET(t(ndmr),7,0.5,"soft","vad")$SigmaY
  #covY are sample covariancematrix of returns by POET,0.5,SOFT,VAD ARE DEFAULTS, 
  #7 FACTORS ARE  RECOMMENDED BY POET PROGRAM,ALTERNATIVE IS ESTIMATE AND FIT
  invcov=ginv(covY)#INVERSE MATRIX PRECISION ESTIMATE
  #SETUP CONSTANTS
  
  
  
  Ahat=t(ones)%*%invcov%*%ones/p
  
  Bhat=t(ones)%*%invcov%*%rbar/p
  Dhat=t(rbar)%*%invcov%*%rbar/p
  opvhat=(Ahat*rho1^2-2*Bhat*rho1+ Dhat)/(Ahat*Dhat-Bhat^2)
  
  w1hat1=matrix(0,p,1)
  w2hat2=matrix(0,p,1)
  w1hat=matrix(0,1,1)
  w2hat=matrix(0,1,1)
  w1hat=(Dhat- rho1*Bhat)/(Ahat*Dhat- Bhat^2)
  w1hat1=(invcov%*%ones/p)%*%w1hat
  w2hat=(rho1*Ahat - Bhat)/(Ahat*Dhat - Bhat^2)
  w2hat2=(invcov%*%rbar/p)%*%w2hat
  wshat=w1hat1+w2hat2#weight in markowitz
  aw=abs(wshat-ws)
  l1e[jj]=colSums(aw)#li error markowitz
  thm3er[jj]=abs(t(wshat)%*%(covY-omz)%*%wshat)
  
  err1[jj]=abs(opvhat/opv-1)
  print(jj)
}


  
  
  
print("variance error")
print(mean(err1))#thm 1 error

print("l1 error of weights")
print(mean(l1e))#thm 2 error

print("thm 3 error:risk")
print(mean(thm3er))

  
  